package de.mdelab.sdm.interpreter.code.debug.ui;

import org.eclipse.core.resources.IMarkerDelta;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IAdapterManager;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.Status;
import org.eclipse.debug.core.DebugEvent;
import org.eclipse.debug.core.DebugException;
import org.eclipse.debug.core.DebugPlugin;
import org.eclipse.debug.core.ILaunch;
import org.eclipse.debug.core.model.IBreakpoint;
import org.eclipse.debug.core.model.IDebugTarget;
import org.eclipse.debug.core.model.IMemoryBlock;
import org.eclipse.debug.core.model.IProcess;
import org.eclipse.debug.core.model.ISourceLocator;
import org.eclipse.debug.core.model.IThread;
import org.eclipse.emf.common.command.AbstractCommand;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.edit.domain.EditingDomain;

import de.mdelab.resourceSetSynchronizer.ISynchronizerCommand;
import de.mdelab.resourceSetSynchronizer.ISynchronizerCommandExecutor;
import de.mdelab.sdm.interpreter.code.debug.ui.breakpoints.SDDebugBreakpoint;
import de.mdelab.sdm.interpreter.code.debug.ui.instanceGraph.SDInstanceGraphEditor;
import de.mdelab.sdm.interpreter.code.debug.ui.stepCommands.IStepBack;
import de.mdelab.sdm.interpreter.code.debug.ui.stepCommands.StepCommandAdapterFactory;
import de.mdelab.sdm.interpreter.code.debug.ui.storyDiagramEditor.SDEditorManager;
import de.mdelab.sdm.interpreter.core.debug.client.SDDebugClient;
import de.mdelab.sdm.interpreter.core.debug.client.stack.DebugClientStackFrame;

public abstract class SDDebugTarget<StoryDiagramElement extends EObject, Activity> extends SDDebugElement<StoryDiagramElement> implements IDebugTarget {
	// private final Logger LOGGER = Logger.getLogger(this.getClass());

	private final ILaunch											launch;
	private final SDDebugEventListenerJob<StoryDiagramElement>		debugEventListenerJob;

// DB: use the first element of the array
//	private final SDDebugThread<StoryDiagramElement, Activity>		debugThread;
	private final SDDebugThread<StoryDiagramElement, Activity>[]	debugThreads;
	private final SDDebugProcess<StoryDiagramElement>				debugProcess;

	private final SDDebugClient<StoryDiagramElement, Activity>		debugClient;

	private SDInstanceGraphEditor									instanceGraphEditor;

	private final SDEditorManager<StoryDiagramElement>				storyDiagramEditorManager;
	
	private final String targetName;
	
	private final int debugClientPort;

	private final int synchronizerAdapterPort;

	@SuppressWarnings("unchecked")
	public SDDebugTarget(	final ILaunch launch,
							final String debugServerAdress,
							final int pi_debugServerPort,
							final int pi_debugClientPort,
							final int pi_synchronizerAdapterPort ) 
	throws CoreException {
		super( null );

		assert launch != null;
		assert debugServerAdress != null;
		assert !"".equals(debugServerAdress);
		assert pi_debugServerPort != pi_synchronizerAdapterPort;

		this.launch = launch;
		
		debugClientPort = pi_debugClientPort;
		synchronizerAdapterPort = pi_synchronizerAdapterPort;
		
		// DB: set more meaningful name related to interpreter server
		targetName = SDDebugUiConstants.SD_DEBUG_TARGET_NAME + " for interpreter running on " + debugServerAdress + ":" + pi_debugServerPort;

		IAdapterManager adapterManager = Platform.getAdapterManager();
		StepCommandAdapterFactory adapterFactory = new StepCommandAdapterFactory();
		adapterManager.registerAdapters(adapterFactory, IStepBack.class);

		/*
		 * Create a new debug client and connect to the debug server
		 */
		debugClient = createSDDebugClient();
		
		// Need to be set before fireCreationEvent since it is the source of creation event
		debugThreads = new SDDebugThread[]{ createSDDebugThread( debugClient ) };

		try	{
			debugClient.connect( debugServerAdress, pi_debugServerPort );
			fireCreationEvent();
		}
		catch ( final Exception p_ex ) {
			p_ex.printStackTrace();
			abort( "Could not connect to debug server '" + debugServerAdress + ":" + pi_debugServerPort + "'.", p_ex );
		}

		configureDebugClient( debugClient );

		/*
		 * Add a source locator to the launch object
		 */
		launch.setSourceLocator(this.createSDDebugSourceLocator());

		/*
		 * Open instance graph editor
		 */
		this.openInstanceGraphEditor();

		this.debugClient.getSynchronizerAdapter().setCommandExecutor(new ISynchronizerCommandExecutor()	{

			@Override
			public void executeCommand(final ISynchronizerCommand command) {
				SDDebugTarget.this.getInstanceGraphEditor().getEditingDomain().getCommandStack().execute(new AbstractCommand() {

					@Override
					public void execute() {
						command.execute();
					}

					@Override
					public void redo() {
						this.execute();
					}

					@Override
					public boolean canUndo() {
						return false;
					}

					@Override
					protected boolean prepare()	{
						return true;
					}
				} );
			}
		} );

		/*
		 * Submit breakpoints to the debugger
		 */
		DebugPlugin.getDefault().getBreakpointManager().addBreakpointListener(this);

		IBreakpoint breakpoints[] = DebugPlugin.getDefault().getBreakpointManager()
				.getBreakpoints(SDDebugUiConstants.DEBUG_MODEL_IDENTIFIER);

		for (IBreakpoint breakpoint : breakpoints)
		{
			assert breakpoint instanceof SDDebugBreakpoint;

			this.addBreakpoint((SDDebugBreakpoint) breakpoint);
		}

		/*
		 * start debug process
		 */
		this.debugProcess = new SDDebugProcess<StoryDiagramElement>(this);

		/*
		 * Create debug client thread.
		 */
		//this.debugThread = this.createSDDebugThread(this.debugClient);
		
		//debugThreads = new SDDebugThread[]	{ createSDDebugThread( debugClient ) };

		/*
		 * Start event listener job
		 */
		this.debugEventListenerJob = new SDDebugEventListenerJob<StoryDiagramElement>(this, this.debugClient);
		this.debugEventListenerJob.schedule();

		/*
		 * Register as breakpoint listener
		 */
		DebugPlugin.getDefault().getBreakpointManager().addBreakpointListener(this);

		/*
		 * Create story diagram manager
		 */
		this.storyDiagramEditorManager = this.createSDEditorManager();

		/*
		 * Send resume command to the client
		 */
		getDebugThread().stepInto();
	}
	
	protected SDDebugThread<StoryDiagramElement, Activity> getDebugThread() {
		return debugThreads[ 0 ];
	}

	protected abstract void configureDebugClient(SDDebugClient<StoryDiagramElement, Activity> debugClient);

	protected abstract SDEditorManager<StoryDiagramElement> createSDEditorManager();

	protected abstract ISourceLocator createSDDebugSourceLocator();

	protected abstract SDDebugClient<StoryDiagramElement, Activity> createSDDebugClient();

	protected abstract SDDebugThread<StoryDiagramElement, Activity> createSDDebugThread(
			SDDebugClient<StoryDiagramElement, Activity> debugClient);

	private void addBreakpoint(SDDebugBreakpoint debugBreakpoint) {
		try {
																			// DB: Handle the skip all breakpoints button
			if ( debugBreakpoint.isEnabled() && debugClient.isConnected() && DebugPlugin.getDefault().getBreakpointManager().isEnabled() ) {
				final String breakpointUUID = debugClient.setBreakpoint( debugBreakpoint.getElementUUID(), debugBreakpoint.getExpressionLanguage(), debugBreakpoint.getExpression() );
				debugBreakpoint.setBreakpointUUID( breakpointUUID );
			}
		}
		catch ( final RuntimeException p_ex ) {
			throw p_ex;
		}
		catch ( final Exception p_ex ) {
			throw new RuntimeException("Could not send breakpoint for element '" + debugBreakpoint.getElementUUID() + "'.", p_ex );
		}
	}

	private void removeBreakpoint(SDDebugBreakpoint debugBreakpoint) {
		try	{
			if (this.debugClient.isConnected())	{
				this.debugClient.removeBreakpoint(debugBreakpoint.getBreakpointUUID());
			}
		}
		catch (Exception e)	{
			throw new RuntimeException("Could not remove breakpoint on element '" + debugBreakpoint.getElementUUID() + "'.", e);
		}
	}

	public SDEditorManager<StoryDiagramElement> getStoryDiagramEditorManager()
	{
		return this.storyDiagramEditorManager;
	}

	public SDDebugClient<StoryDiagramElement, Activity> getDebugClient() {
		return this.debugClient;
	}

	@Override
	public boolean canTerminate() {
		return getDebugThread().canTerminate();
	}

	@Override
	public boolean isTerminated() {
		return getDebugThread().isTerminated();
	}

	@Override
	public void fireTerminateEvent() {

		// DB: Only remove highlighting instead of closing the diagram editor
//				this.closeInstanceGraphEditor();
		//
//				this.storyDiagramEditorManager.closeAllStoryDiagramEditors();
		storyDiagramEditorManager.clearHighlighting();

		DebugPlugin.getDefault().getBreakpointManager().removeBreakpointListener( this );

		super.fireTerminateEvent();
	}

	public void debuggerTerminated() 
	throws DebugException {
		getDebugThread().terminateClient( false );
		
		fireTerminateEvent();
	}

	public void debuggerResumed() {
		getDebugThread().setSuspended( false );

		fireResumeEvent( DebugEvent.UNSPECIFIED );
	}

	@Override
	public void fireResumeEvent( final int pi_detail )	{
		instanceGraphEditor.setReadOnly( true );
		
		super.fireResumeEvent( pi_detail );
	}

	@Override
	public void terminate()
	throws DebugException {
		
		System.out.println( "Terminating debug target..." );

//		if ( !getDebugThread().isTerminated() ) {
		getDebugThread().terminate();
//		}

		// DB: Not needed because this flag is set by the terminate method
		//this.debugThread.setTerminated();
		// now called from debug thread
		//fireTerminateEvent();
	}

	@Override
	public boolean canResume() {
		return getDebugThread().canResume();
	}

	@Override
	public boolean canSuspend()	{
		return getDebugThread().canSuspend();
	}

	@Override
	public boolean isSuspended() {
		return getDebugThread().isSuspended();
	}

	@Override
	public void resume()
	throws DebugException {
		getDebugThread().resume();

		instanceGraphEditor.setReadOnly( true );
	}
	
	protected void handleException( final String p_message,
									final Throwable p_th,
									final boolean pb_abort )
	throws DebugException {
		if ( pb_abort ) {
			terminate();
		}
		
		if ( p_th instanceof DebugException ) {
			throw (DebugException) p_th;
		}
		
		throw new DebugException( new Status( IStatus.ERROR, SDDebugUiPlugin.getDefault().getBundle().getSymbolicName(), p_message, p_th ) );
	}
	
	private void doDebuggerSuspended() {
		getDebugThread().setSuspended( true );

		instanceGraphEditor.setReadOnly( false );

		/*
		 * Open the story diagram that contains the current element.
		 */
		try	{
			getDebugThread().updateStackFrames();

			final DebugClientStackFrame<StoryDiagramElement> stackFrame = debugClient.getExecutionStack().peek();
			storyDiagramEditorManager.highlightStoryDiagramElement(	stackFrame.getStoryDiagramElement(), stackFrame.getLastNotificationType() );
		}
		catch ( Throwable p_th ) {
			try {
				terminate();
			} 
			catch ( final DebugException p_ex ) {
				p_ex.printStackTrace();
			}
			
			throw new RuntimeException( "Could not highlight current story diagram element.", p_th );
		}
	}

	public void debuggerSuspended() {
		doDebuggerSuspended();

		getDebugThread().fireSuspendEvent( DebugEvent.UNSPECIFIED );
	}

	public void debuggerSuspendedAtBreakpoint() {
		doDebuggerSuspended();
		
		getDebugThread().fireSuspendEvent( DebugEvent.BREAKPOINT );
	}

	@Override
	public void suspend()
	throws DebugException {
		getDebugThread().suspend();
		
		fireSuspendEvent( DebugEvent.UNSPECIFIED );
	}
	
	@Override
	public void breakpointAdded(IBreakpoint breakpoint) {
		assert breakpoint instanceof SDDebugBreakpoint;

		this.addBreakpoint((SDDebugBreakpoint) breakpoint);

		fireChangeEvent( DebugEvent.BREAKPOINT );
	}

	@Override
	public void breakpointRemoved(IBreakpoint breakpoint, IMarkerDelta delta) {
		assert breakpoint instanceof SDDebugBreakpoint;

		removeBreakpoint((SDDebugBreakpoint) breakpoint);
		
		fireChangeEvent( DebugEvent.BREAKPOINT );
	}

	@Override
	public void breakpointChanged(IBreakpoint breakpoint, IMarkerDelta delta) {
		assert breakpoint instanceof SDDebugBreakpoint;

		removeBreakpoint((SDDebugBreakpoint) breakpoint);

		addBreakpoint((SDDebugBreakpoint) breakpoint);

		fireChangeEvent( DebugEvent.BREAKPOINT );
	}

	@Override
	public boolean canDisconnect() {
		return false;
	}

	@Override
	public void disconnect()
	throws DebugException {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean isDisconnected()	{
		return false;
	}

	@Override
	public boolean supportsStorageRetrieval() {
		return false;
	}

	@Override
	public IMemoryBlock getMemoryBlock(long startAddress, long length) throws DebugException
	{
		throw new UnsupportedOperationException();
	}

	@Override
	public IProcess getProcess()
	{
		return this.debugProcess;
	}

	@Override
	public IThread[] getThreads()
	throws DebugException {
		return debugThreads;
	}

	@Override
	public boolean hasThreads()
	throws DebugException {
		return debugThreads.length > 0;
	}

	@Override
	public String getName()
	throws DebugException {
		return targetName;//SDDebugUiConstants.SD_DEBUG_TARGET_NAME;
	}

	@Override
	public boolean supportsBreakpoint(IBreakpoint breakpoint) {
		return breakpoint.getModelIdentifier().equals(SDDebugUiConstants.DEBUG_MODEL_IDENTIFIER);
	}

	@Override
	public SDDebugTarget<StoryDiagramElement, ?> getDebugTarget() {
		return this;
	}

	@Override
	public ILaunch getLaunch() {
		return launch;
	}

//	@Override
//	public void fireResumeEvent( final int pi_detail )	{
//		//getDebugThread().setSuspended(false);
//
//		this.instanceGraphEditor.setReadOnly(true);
//		
//		// DB: set the top debug thread as the source of the event instead of this debug target in order to select the correct element on the stack frame tree of debuger view
//		fireEvent( DebugEvent.RESUME, pi_detail );
////		super.fireResumeEvent(detail);
//	}

//	@Override
//	public void fireSuspendEvent( int pi_detail ) {
//		getDebugThread().setSuspended( true );
//
//		instanceGraphEditor.setReadOnly( false );
//
//		/*
//		 * Open the story diagram that contains the current element.
//		 */
//		try	{
//			getDebugThread().updateStackFrames();
//
//			final DebugClientStackFrame<StoryDiagramElement> stackFrame = debugClient.getExecutionStack().peek();
//			this.storyDiagramEditorManager.highlightStoryDiagramElement(	stackFrame.getStoryDiagramElement(),
//																			stackFrame.getLastNotificationType() );
//
//			// DB: set the debug thread as the source of the event instead of this debug target so that the top element of the stack frame tree of 
//			// the debugger view is selected after each step
//			//fireEvent( DebugEvent.SUSPEND, pi_detail );
//			//super.fireSuspendEvent(detail);
//		}
//		catch ( Throwable p_th ) {
//			throw new RuntimeException("Could not highlight current story diagram element.", p_th );
//		}
//	}

//	@Override
//	public void fireTerminateEvent() {
//		System.out.println( "Firing terminated event..." );
//
//		super.fireTerminateEvent();
//		
//		getDebugThread().fireTerminateEvent();
		
//		fireEvent( DebugEvent.TERMINATE );
//
		// DB: fire moved to terminate so that when user terminates debugging from the tool bar button
		// the UI elements are also notified
//		try {
//			this.terminate();
//		}
//		catch ( final DebugException p_ex ) {
//			throw new RuntimeException( p_ex );
//		}
//	}

	protected abstract void openInstanceGraphEditor();

	protected abstract void closeInstanceGraphEditor();

	protected abstract SDInstanceGraphEditor createSDInstanceGraphEditor();

	public SDInstanceGraphEditor getInstanceGraphEditor() {
		return instanceGraphEditor;
	}

	public void setInstanceGraphEditor(SDInstanceGraphEditor instanceGraphEditor)
	{
		this.instanceGraphEditor = instanceGraphEditor;
	}

	public EditingDomain getDebuggerEditingDomain()	{
		return instanceGraphEditor.getEditingDomain();
	}
	
	protected int getDebugClientPort() {
		return debugClientPort;
	}

	protected int getSynchronizerAdapterPort() {
		return synchronizerAdapterPort;
	}
}
