package de.mdelab.sdm.interpreter.code.debug.ui;

import java.util.LinkedList;
import java.util.List;
import java.util.Stack;

import org.eclipse.debug.core.DebugEvent;
import org.eclipse.debug.core.DebugException;
import org.eclipse.debug.core.model.IBreakpoint;
import org.eclipse.debug.core.model.IStackFrame;
import org.eclipse.debug.core.model.IThread;
import org.eclipse.emf.ecore.EObject;

import de.mdelab.sdm.interpreter.core.debug.client.SDDebugClient;
import de.mdelab.sdm.interpreter.core.debug.client.stack.DebugClientStackFrame;

public abstract class SDDebugThread<StoryDiagramElement extends EObject, Activity> extends SDDebugElement<StoryDiagramElement> implements IThread {

	private final SDDebugClient<StoryDiagramElement, Activity>	debugClient;

	private boolean												terminated;

	private boolean												suspended;

	private SDDebugUiStackFrame<StoryDiagramElement>[]			stackFrames;

	public SDDebugThread( 	final SDDebugTarget<StoryDiagramElement, Activity> p_target,
							final SDDebugClient<StoryDiagramElement, Activity> p_debugClient ) {
		super( p_target );

		assert p_debugClient != null;

		debugClient = p_debugClient;

		terminated = false;

		setSuspended( false );
	}

	@Override
	public boolean canResume() {
		return !isTerminated() && isSuspended();
	}

	@Override
	public boolean canSuspend()	{
		return !isTerminated() && !isSuspended();
	}

	@Override
	public boolean isSuspended() {
		return suspended;
	}

	@Override
	public void resume()
	throws DebugException {
		try	{
			debugClient.resume();
			
			setSuspended( false );
		}
		catch ( final Throwable p_th ) {
			handleException( "Could not resume debugger.", p_th, true );
		}
	}

	@Override
	public void suspend()
	throws DebugException {
		try	{
			debugClient.suspend();

			setSuspended( true );
		}
		catch ( final Throwable p_th ) {
			handleException( "Could not suspend debugger.", p_th, true );
		}
	}

	@Override
	public boolean canStepInto() {
		return isSuspended();
	}

	@Override
	public boolean canStepOver() {
		return isSuspended();
	}

	@Override
	public boolean canStepReturn() {
		return isSuspended();
	}

	@Override
	public boolean isStepping() {
		return !isSuspended() && !isTerminated();
	}

	@Override
	public void stepInto()
	throws DebugException {
		try	{
			this.debugClient.stepInto();
		}
		catch ( final Throwable p_th ) {
			handleException( "Could not step into.", p_th, true );
		}
	}

	@Override
	public void stepOver()
	throws DebugException {
		try	{
			debugClient.stepOver();
		}
		catch ( final Throwable p_th ) {
			handleException( "Could not step over.", p_th, true );
		}
	}

	@Override
	public void stepReturn()
	throws DebugException {
		try	{
			debugClient.stepOut();
		}
		catch ( final Throwable p_th ) {
			handleException( "Could not step out.", p_th, true );
		}
	}

	public void stepBackOver()
	throws DebugException {
		try	{
			debugClient.stepBackOver();
		}
		catch ( final Throwable p_th ) {
			handleException( "Could not step back over.", p_th, true );
		}
	}

	@Override
	public boolean canTerminate() {
		return !isTerminated();
	}

	@Override
	public boolean isTerminated() {
		return terminated;
	}
	
	public void terminateClient( final boolean pb_deconnectRemote ) 
	throws DebugException {
		terminated = true;
		setSuspended( false );

		debugClient.disconnect( pb_deconnectRemote );
		fireTerminateEvent();
	}

	@Override
	public void terminate()
	throws DebugException {
		System.out.println( "Terminating thread..." );
		
		try	{
			debugClient.terminate();
		}
		catch ( final Throwable p_th ) {
			handleException( "Could not terminate the debugger.", p_th, false );
		}
		finally {
			terminateClient( true );
			getDebugTarget().fireTerminateEvent();
		}
//		
//		terminated = true;
//		setSuspended( false );
//
//		try	{
//			this.debugClient.terminate();
//			this.debugClient.disconnect();
//
//			fireTerminateEvent();
//		}
//		catch ( final Throwable p_th ) {
//			handleException( "Could not terminate the debugger.", p_th, false );
//		}
	}
	
	private void handleException( 	final String p_message,
									final Throwable p_th,
									final boolean pb_abort )
	throws DebugException {
		getDebugTarget().handleException( p_message, p_th, pb_abort );
	}

	@Override
	public IStackFrame[] getStackFrames()
	throws DebugException {
		if ( stackFrames == null ) {
			updateStackFrames();
		}

		return stackFrames;
	}

	@SuppressWarnings("unchecked")
	public void updateStackFrames()
	throws DebugException {
		try	{
			final Stack<DebugClientStackFrame<StoryDiagramElement>> stack = this.debugClient.getExecutionStack();

			final List<SDDebugUiStackFrame<StoryDiagramElement>> stackFramesList = new LinkedList<SDDebugUiStackFrame<StoryDiagramElement>>();

			for ( final DebugClientStackFrame<StoryDiagramElement> stackFrame : stack ) {
				stackFramesList.add( 0, createSDDebugUiStackFrame( (SDDebugTarget<StoryDiagramElement, Activity>) getDebugTarget(), this, stackFrame ) );
			}

			stackFrames = stackFramesList.toArray( new SDDebugUiStackFrame[ stackFramesList.size() ] );
		}
		catch ( final Throwable p_th ) {
			handleException( "Error while querying execution stack.", p_th, true );
		}
	}

	protected abstract SDDebugUiStackFrame<StoryDiagramElement> createSDDebugUiStackFrame( 	SDDebugTarget<StoryDiagramElement, Activity> p_sdDebugTarget,
																							SDDebugThread<StoryDiagramElement, Activity> p_sdDebugThread,
																							DebugClientStackFrame<StoryDiagramElement> p_stackFrame );

	@Override
	public boolean hasStackFrames()
	throws DebugException {
		return !isTerminated();
	}

	@Override
	public int getPriority()
	throws DebugException {
		return 0;
	}

	@Override
	public IStackFrame getTopStackFrame()
	throws DebugException {
		try {
			return getStackFrames()[ 0 ];
		}
		catch ( final Throwable p_th ) {
			handleException( "Error while querying execution stack.", p_th, true );
			
			return null;
		}
	}

	@Override
	public String getName()
	throws DebugException {
		return SDDebugUiConstants.SD_INTERPRETER_THREAD_NAME;
	}

	@Override
	public IBreakpoint[] getBreakpoints() {
		if ( !isSuspended() ) {
			throw new UnsupportedOperationException();
		}

		return null;
	}

	protected void setSuspended( final boolean pb_suspended )	{
		suspended = pb_suspended;
		
		if ( isSuspended() ) {
			fireSuspendEvent( DebugEvent.UNSPECIFIED );
		}
		else {
			fireResumeEvent( DebugEvent.UNSPECIFIED );
		}
	}
}
