package de.mdelab.sdm.interpreter.core.debug;

import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.Map;
import java.util.Stack;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;

import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.change.ChangeDescription;
import org.eclipse.emf.ecore.change.util.ChangeRecorder;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.net4j.Net4jUtil;
import org.eclipse.net4j.connector.IConnector;
import org.eclipse.net4j.signal.SignalProtocol;
import org.eclipse.net4j.tcp.TCPUtil;
import org.eclipse.net4j.util.container.ContainerUtil;
import org.eclipse.net4j.util.container.IManagedContainer;
import org.eclipse.net4j.util.event.IEvent;
import org.eclipse.net4j.util.event.IListener;
import org.eclipse.net4j.util.lifecycle.ILifecycleEvent;
import org.eclipse.net4j.util.lifecycle.LifecycleUtil;
import org.eclipse.spi.net4j.ServerProtocolFactory;

import de.mdelab.resourceSetSynchronizer.EObjectUuidCalculator;
import de.mdelab.resourceSetSynchronizer.SynchronizerAdapter;
import de.mdelab.sdm.interpreter.core.AbortInterpretationException;
import de.mdelab.sdm.interpreter.core.SDMInterpreter;
import de.mdelab.sdm.interpreter.core.debug.breakpoints.Breakpoint;
import de.mdelab.sdm.interpreter.core.debug.breakpoints.StepBreakpoint;
import de.mdelab.sdm.interpreter.core.debug.breakpoints.StepIntoBreakpoint;
import de.mdelab.sdm.interpreter.core.debug.breakpoints.StepOutBreakpoint;
import de.mdelab.sdm.interpreter.core.debug.breakpoints.StepOverBreakpoint;
import de.mdelab.sdm.interpreter.core.debug.protocol.SDDebugProtocolCommandIndicationFactory;
import de.mdelab.sdm.interpreter.core.debug.protocol.requests.ExecutionFinishedEventRequest;
import de.mdelab.sdm.interpreter.core.debug.protocol.requests.ExecutionResumedEventRequest;
import de.mdelab.sdm.interpreter.core.debug.protocol.requests.ExecutionSuspendedEventRequest;
import de.mdelab.sdm.interpreter.core.debug.protocol.requests.ExecutionTerminatedEventRequest;
import de.mdelab.sdm.interpreter.core.facade.EForEachSemantics;
import de.mdelab.sdm.interpreter.core.notifications.ActivityNodeExecutionStartedNotification;
import de.mdelab.sdm.interpreter.core.notifications.InstanceLinkDestroyedNotification;
import de.mdelab.sdm.interpreter.core.notifications.InstanceObjectCreatedNotification;
import de.mdelab.sdm.interpreter.core.notifications.InterpreterNotification;
import de.mdelab.sdm.interpreter.core.notifications.NotificationReceiver;
import de.mdelab.sdm.interpreter.core.notifications.NotificationTypeEnum;
import de.mdelab.sdm.interpreter.core.notifications.VariableCreatedNotification;
import de.mdelab.sdm.interpreter.core.notifications.VariableDeletedNotification;
import de.mdelab.sdm.interpreter.core.notifications.VariableValueChangedNotification;
import de.mdelab.sdm.interpreter.core.variables.Variable;
import de.mdelab.sdm.interpreter.core.variables.VariablesScope;

public abstract class SDDebugger<StoryDiagramElement extends EObject, Activity extends EObject, ActivityNode, StoryPattern, Expression>
		implements NotificationReceiver<EClassifier> {
	
	private EDebuggerState																								debuggerState;

	private final IManagedContainer																						debugProtocolContainer;

	private final int																									debugServerPort;

	private final int																									resourceSetSynchronizerPort;

	private final SignalProtocol<Object>																				debugProtocol;

	private final Map<String, Collection<Breakpoint<StoryDiagramElement>>>												breakpoints;

	private final Stack<StackFrame<StoryDiagramElement>>																executionStack;

	private final Stack<ChangeHistoryEntry>																				changeHistory;

	private ResourceSet																									debuggerResourceSet;

	public static final URI																								DEBUGGER_RESOURCE_URI	= URI.createURI("de.mdelab.sdm.interpreter.sde.debug.resource");

	/*
	 * The debugger resource contains all objects that are not contained in any
	 * other resource.
	 */
	private Resource																									debuggerResource;

	private ChangeRecorder																								changeRecorder;

	private SynchronizerAdapter																							synchronizerAdapter;

	private SDMInterpreter<Activity, ActivityNode, ?, StoryPattern, ?, ?, EClassifier, EStructuralFeature, Expression>	interpreter;

	private boolean																										processNotifications;

	private AbortInterpretationException																				abortException;
	
	private boolean initialized;

	@SuppressWarnings("unchecked")
	public SDDebugger( 	final SDMInterpreter<Activity, ActivityNode, ?, StoryPattern, ?, ?, EClassifier, EStructuralFeature, Expression> interpreter,
						final int pi_debugServerPort,
						final int pi_resourceSetSynchronizerPort ) {
		assert interpreter != null;
		
		initialized = false;

		this.interpreter = interpreter;
		interpreter.getNotificationEmitter().addNotificationReceiver(this);

		this.processNotifications = true;

		this.debugServerPort = pi_debugServerPort;
		this.resourceSetSynchronizerPort = pi_resourceSetSynchronizerPort;// + 1;

		this.debuggerState = EDebuggerState.SUSPENDED;

		this.breakpoints = new ConcurrentHashMap<String, Collection<Breakpoint<StoryDiagramElement>>>();

		this.executionStack = new Stack<StackFrame<StoryDiagramElement>>();
		this.executionStack.push(new StackFrame<StoryDiagramElement>());

		this.changeHistory = new Stack<ChangeHistoryEntry>();

		/*
		 * Open debug server
		 */
		this.debugProtocolContainer = ContainerUtil.createContainer();

		Net4jUtil.prepareContainer(this.debugProtocolContainer);
		TCPUtil.prepareContainer(this.debugProtocolContainer);
		this.debugProtocolContainer.registerFactory(this.createDebugProtocolCommandIndicationFactory());

		LifecycleUtil.activate(this.debugProtocolContainer);

		TCPUtil.getAcceptor(this.debugProtocolContainer, "0.0.0.0:" + this.debugServerPort);

		this.debugProtocol = (SignalProtocol<Object>) debugProtocolContainer.getFactory(ServerProtocolFactory.PRODUCT_GROUP,
				SDDebugProtocolCommandIndicationFactory.SDE_DEBUG_PROTOCOL).create(null);

		/*
		 * Attach listener to the debug protocol to be notified when the
		 * connection is closed
		 */
		this.debugProtocol.addListener(new IListener() {
			
			@Override
			public void notifyEvent(IEvent event) {
				if (event instanceof ILifecycleEvent) {
					if (((ILifecycleEvent) event).getKind() == ILifecycleEvent.Kind.DEACTIVATED) {
						
						try	{
							/*
							 * TODO Avoid this hack!
							 * 
							 * Start a new thread that calls disconnect after
							 * one second. This is necessary to avoid calling
							 * disconnect from within this event handler.
							 */
							new Thread() {
								@Override
								public void run() {
									try {
										// DB: this does not seem to be needed...
										//Thread.sleep(1000);

										terminate( false );

										disconnect( false );
									}
									catch (Exception e)	{
										e.printStackTrace();
									}
								};
							}.start();
						}
						catch (Exception e)	{
							e.printStackTrace();
							
							throw new RuntimeException(e);
						}
					}
				}
			}
		});
	}

	protected abstract SDDebugProtocolCommandIndicationFactory<StoryDiagramElement> createDebugProtocolCommandIndicationFactory();

	protected void initialize(//Activity activity,
								final ResourceSet p_resSet/*,
								final SDMInterpreter<Activity, ActivityNode, ?, StoryPattern, ?, ?, EClassifier, EStructuralFeature, Expression> interpreter*/)
	{
		//assert activity.eResource() != null;
		assert p_resSet /*activity.eResource().getResourceSet()*/ != null;

		// DB: Issue 1825: Interpreter is already set in constructor
//		assert interpreter != null;
//
//		this.interpreter = interpreter;

		/*
		 * Get the resource set of the activity
		 */
		this.debuggerResourceSet = p_resSet;//activity.eResource().getResourceSet();

		/*
		 * Check if the debugger resource already exists
		 */
		this.debuggerResource = this.debuggerResourceSet.getResource(SDDebugger.DEBUGGER_RESOURCE_URI, false);

		if (this.debuggerResource != null)
		{
			/*
			 * Clear the existing debugger resource
			 */
			this.debuggerResource.getContents().clear();
		}
		else
		{
			/*
			 * create a new debugger resource
			 */
			this.debuggerResource = this.debuggerResourceSet.createResource(SDDebugger.DEBUGGER_RESOURCE_URI);
		}

		/*
		 * Attach the resource set synchronizer to the resource set
		 */
		this.synchronizerAdapter = new SynchronizerAdapter(this.debuggerResourceSet, this.resourceSetSynchronizerPort);
		this.synchronizerAdapter.getEObjectUUIDManager().setEObjectUuidCalculator(this.getStoryDiagramElementUuidCalculator());

		/*
		 * Attach the change recorder to the debugger resource set
		 */
		this.changeRecorder = new ChangeRecorder(this.debuggerResourceSet);
		
		initialized = true;
	}

	public boolean isInitialized() {
		return initialized;
	}

	protected abstract EObjectUuidCalculator getStoryDiagramElementUuidCalculator();

	private synchronized void disconnect( final boolean pb_deconnectRemote )
	throws Exception {
		if ( debugProtocolContainer.isActive() )	{
			System.out.println("Shutting down server...");

			this.abort(null);

			this.debugProtocolContainer.deactivate();

			this.synchronizerAdapter.shutdown( pb_deconnectRemote );

			System.out.println("Server shut down.");

			assert !this.debugProtocolContainer.isActive();
			assert !this.debugProtocol.isActive();
		}
	}

	public void addBreakpoint(Breakpoint<StoryDiagramElement> breakpoint) {
		Collection<Breakpoint<StoryDiagramElement>> bp = this.breakpoints.get(breakpoint.getElementUUID());

		if (bp == null)
		{
			bp = new ConcurrentLinkedQueue<Breakpoint<StoryDiagramElement>>();
			this.breakpoints.put(breakpoint.getElementUUID(), bp);
		}

		bp.add(breakpoint);
	}

	public void removeBreakpoint(String breakpointUUID)
	{
		assert breakpointUUID != null;

		/*
		 * Search for the breakpoint with this UUID
		 */
		for (Iterator<Collection<Breakpoint<StoryDiagramElement>>> it1 = this.getBreakpoints().values().iterator(); it1.hasNext();)
		{
			Collection<Breakpoint<StoryDiagramElement>> breakpoints = it1.next();

			for (Iterator<Breakpoint<StoryDiagramElement>> it = breakpoints.iterator(); it.hasNext();)
			{
				if (it.next().getBreakpointUUID().equals(breakpointUUID))
				{
					it.remove();
				}
			}

			if (breakpoints.isEmpty())
			{
				it1.remove();
			}
		}
	}

	public void suspendOnBreakpoints(StoryDiagramElement currentlyExecutedElement,
			VariablesScope<?, ?, ?, ?, ?, ?, EClassifier, ?, ?> variablesScope)
	{
		assert currentlyExecutedElement != null;
		assert variablesScope != null;

		synchronized (this)
		{
			if (this.abortException != null)
			{
				try
				{
					System.out.println("ABORTING INTERPRETER...");

					throw this.abortException;
				}
				finally
				{
					this.abortException = null;
				}
			}

			/*
			 * Check each breakpoint of the currently executed element
			 */
			Collection<Breakpoint<StoryDiagramElement>> bp = breakpoints.get( getStoryDiagramElementUuidCalculator().calculateUuid( currentlyExecutedElement ) );

			Breakpoint<StoryDiagramElement> breakpoint = null;

			if (bp != null
					&& (breakpoint = this.evaluateBreakpoints(currentlyExecutedElement, variablesScope, bp)) != null
					|| this.breakpoints.get(Breakpoint.ANY_ELEMENT_UUID) != null
					&& (breakpoint = this.evaluateBreakpoints(currentlyExecutedElement, variablesScope,
							this.breakpoints.get(Breakpoint.ANY_ELEMENT_UUID))) != null) {
				debuggerState = EDebuggerState.SUSPENDED;

				System.out.println("SUSPENDING INTERPRETER...");

				try	{
					if (breakpoint instanceof StepBreakpoint) {
						new ExecutionSuspendedEventRequest(this.debugProtocol, null).sendAsync();
					}
					else {
						new ExecutionSuspendedEventRequest(this.debugProtocol, breakpoint.getBreakpointUUID()).sendAsync();
					}

					this.wait();
				}
				catch (Exception e)	{
					e.printStackTrace();
					
					throw new RuntimeException(e);
				}

				System.out.println("RESUMING INTERPRETER...");

				if ( abortException != null  ) {
					try	{
						System.out.println("ABORTING INTERPRETER...");

						throw abortException;
					}
					finally {
						this.abortException = null;
					}
				}
			}
		}
	}

	protected Breakpoint<StoryDiagramElement> evaluateBreakpoints(StoryDiagramElement currentlyExecutedElement,
			VariablesScope<?, ?, ?, ?, ?, ?, EClassifier, ?, ?> variablesScope, Collection<Breakpoint<StoryDiagramElement>> breakpoints)
	{
		Iterator<Breakpoint<StoryDiagramElement>> it = breakpoints.iterator();

		while (it.hasNext())
		{
			Breakpoint<StoryDiagramElement> breakpoint = it.next();

			if (breakpoint.evaluate(currentlyExecutedElement, variablesScope))
			{
				if (breakpoint instanceof StepBreakpoint)
				{
					it.remove();
				}

				return breakpoint;
			}
		}

		return null;
	}

	/**
	 * Resumes the debugger thread.
	 * 
	 * @throws Exception
	 */
	public void resume() 
	throws Exception {
		synchronized ( this ) {
			if ( debuggerState != EDebuggerState.RUNNING ) {
				debuggerState = EDebuggerState.RUNNING;

				// DB: We need a synchronous request otherwise under some conditions when stepping into,
				// the suspend request may be processed by the client before the resume request thus causing 
				// the debugger client to enter a resumed state while the server is in suspend state
				if ( new ExecutionResumedEventRequest( debugProtocol ).send() ) {
					notify();
				}
			}
		}
	}

	public void stepInto() 
	throws Exception {
		if ( getDebuggerState() == EDebuggerState.SUSPENDED ) {
			addBreakpoint( new StepIntoBreakpoint<StoryDiagramElement>( this ) );
			
			resume();
		}
	}

	public void stepOver()
	throws Exception {
		if ( getDebuggerState() == EDebuggerState.SUSPENDED ) {
			addBreakpoint( new StepOverBreakpoint<StoryDiagramElement>( this ) );
			
			resume();
		}
	}

	public void stepOut() throws Exception
	{
		if (this.getDebuggerState() == EDebuggerState.SUSPENDED)
		{
			this.addBreakpoint(new StepOutBreakpoint<StoryDiagramElement>(this));
			this.resume();
		}
	}

	@SuppressWarnings("unchecked")
	public void stepBackOver() throws Exception
	{
		if (this.getDebuggerState() == EDebuggerState.SUSPENDED && this.getChangeHistory().size() > 2)
		{
			this.setProcessNotifications(false);

			/*
			 * Stop recording changes and revert the last change
			 */
			this.changeRecorder.endRecording().apply();

			/*
			 * Revert all changes until the topmost notification is starting
			 * activity node execution.
			 */
			int currentStackSize = this.getExecutionStack().size();

			do
			{
				/*
				 * Pop the last change description
				 */
				this.revertLastChangeHistoryEntry();

			}
			while (!this.getChangeHistory().isEmpty()
					&& (this.getChangeHistory().peek().getInterpreterNotification().getNotificationType() != NotificationTypeEnum.ACTIVITY_NODE_EXECUTION_STARTED || currentStackSize < this
							.getExecutionStack().size()));

			/*
			 * Also revert the starting of the activity node and set this
			 * activity node as the next executed node.
			 */
			ChangeHistoryEntry changeEntry = this.revertLastChangeHistoryEntry();

			assert changeEntry.getInterpreterNotification().getNotifier() instanceof SDMInterpreter;

			/*
			 * Set the next node to be executed by the interpreter.
			 */
			((SDMInterpreter<?, ActivityNode, ?, ?, ?, ?, ?, ?, ?>) changeEntry.getInterpreterNotification().getNotifier())
					.setNextNode(((ActivityNodeExecutionStartedNotification<ActivityNode, ?>) changeEntry.getInterpreterNotification())
							.getActivityNode());

			this.addBreakpoint(new StepIntoBreakpoint<StoryDiagramElement>(this));

			this.changeRecorder.beginRecording(Collections.singleton(this.debuggerResourceSet));

			this.setProcessNotifications(true);

			this.abort((SDMInterpreter<?, ?, ?, ?, ?, ?, ?, ?, ?>) changeEntry.getInterpreterNotification().getNotifier());
		}
	}

	@SuppressWarnings("unchecked")
	protected ChangeHistoryEntry revertLastChangeHistoryEntry()
	{
		ChangeHistoryEntry changeEntry = this.getChangeHistory().pop();

		System.out.println("Reverting execution of '" + changeEntry.getInterpreterNotification() + "...'");

		/*
		 * Undo changes on the instance model
		 */
		changeEntry.getChangeDescription().apply();

		/*
		 * If the last change was one of the following (finishing the execution
		 * of something), push a new stack frame onto the execution stack.
		 */
		switch ( changeEntry.getInterpreterNotification().getNotificationType() ) {
			case ACTIVITY_EXECUTION_FINISHED:
			case ACTIVITY_NODE_EXECUTION_FINISHED:
			case EVALUATED_EXPRESSION:
			case STORY_PATTERN_INITIALIZATION_FINISHED:
			case STORY_PATTERN_MATCHING_SUCCESSFUL:
			case STORY_PATTERN_MATCHING_FAILED:
			case STORY_PATTERN_APPLICATION_FINISHED:
				getExecutionStack().push( new StackFrame<StoryDiagramElement>() );
				
				break;
			
			default:
				break;
		}

		/*
		 * If the previous change was starting the execution of something, pop
		 * the topmost element of the execution stack.
		 */
		switch ( getChangeHistory().peek().getInterpreterNotification().getNotificationType() )	{
			case ACTIVITY_EXECUTION_STARTED:
			case ACTIVITY_NODE_EXECUTION_STARTED:
			case EVALUATING_EXPRESSION:
			case STORY_PATTERN_INITIALIZATION_STARTED:
			case STORY_PATTERN_MATCHING_STARTED:
			case STORY_PATTERN_APPLICATION_STARTED:
				getExecutionStack().pop();
				
				break;
			
			default:
				break;
		}

		/*
		 * Find the last change in the change history that was connected to a
		 * story diagram element. Modify the execution stack accordingly.
		 */
		ChangeHistoryEntry previousChange = null;

		for (int i = this.changeHistory.size() - 1; i >= 0; i--)
		{
			previousChange = this.changeHistory.get(i);

			if (previousChange.getInterpreterNotification().getMainStoryDiagramElement() != null)
			{
				StackFrame<StoryDiagramElement> stackFrame = this.getExecutionStack().peek();

				stackFrame.setStoryDiagramElement((StoryDiagramElement) previousChange.getInterpreterNotification()
						.getMainStoryDiagramElement());
				stackFrame.setVariablesScope(previousChange.getInterpreterNotification().getVariablesScope());
				stackFrame.setLastNotification(previousChange.getInterpreterNotification());
				break;
			}
		}

		/*
		 * Revert changes of variable scopes.
		 */
		switch ( changeEntry.getInterpreterNotification().getNotificationType() ) {
			case VARIABLE_CREATED:
				changeEntry
						.getInterpreterNotification()
						.getVariablesScope()
						.deleteVariable(
								((VariableCreatedNotification<EClassifier>) changeEntry.getInterpreterNotification()).getVariable()
										.getName());
				
				break;

			case VARIABLE_DELETED:
				Variable<EClassifier> variable = ((VariableDeletedNotification<EClassifier>) changeEntry.getInterpreterNotification())
						.getVariable();
				changeEntry.getInterpreterNotification().getVariablesScope()
						.createVariable(variable.getName(), variable.getClassifier(), variable.getValue());
				
				break;

			case VARIABLE_VALUE_CHANGED:
				VariableValueChangedNotification<EClassifier> n = (VariableValueChangedNotification<EClassifier>) changeEntry
						.getInterpreterNotification();
				changeEntry.getInterpreterNotification().getVariablesScope()
						.changeVariableValue(n.getVariable().getName(), n.getOldValue());
				
				break;
			
			default:
				break;
		}

		/*
		 * Revert changes of the interpreter's internal state, especially
		 * pattern matchers for for-each nodes
		 */
		switch ( changeEntry.getInterpreterNotification().getNotificationType() ) {
			case STORY_PATTERN_INITIALIZATION_STARTED:
				assert this.changeHistory.peek().getInterpreterNotification() instanceof ActivityNodeExecutionStartedNotification<?, ?>;

				if (this.interpreter
						.getFacadeFactory()
						.getStoryNodeFacade()
						.getForEachSemantics(
								((ActivityNodeExecutionStartedNotification<ActivityNode, EClassifier>) this.changeHistory.peek()
										.getInterpreterNotification()).getActivityNode()) == EForEachSemantics.FRESH_MATCH)
				{
					assert this.changeHistory.peek().getInterpreterNotification().getNotifier() instanceof SDMInterpreter;

					((SDMInterpreter<Activity, ActivityNode, ?, StoryPattern, ?, ?, EClassifier, EStructuralFeature, Expression>) this.changeHistory
							.peek().getInterpreterNotification().getNotifier()).getStoryPatternMatchers().remove(
							changeEntry.getInterpreterNotification().getNotifier());
				}
				
				break;
			
			default:
				break;
		}

		return changeEntry;
	}

	/**
	 * Suspends the debugger by creating a step into breakpoint. This implies
	 * that the debugger is not immediately suspended but when breakpoints are
	 * checked next time.
	 * 
	 * @throws Exception
	 */
	public void suspend() throws Exception
	{
		if (this.getDebuggerState() == EDebuggerState.RUNNING)
		{
			this.addBreakpoint(new StepIntoBreakpoint<StoryDiagramElement>(this));
		}
	}

	public EDebuggerState getDebuggerState()
	{
		return this.debuggerState;
	}

	/**
	 * Terminates the debugger.
	 * 
	 * @throws Exception
	 */
	public void terminate( final boolean pb_deconnectRemote )
	throws Exception {
		if (this.debuggerState != EDebuggerState.TERMINATED) {
			this.abort(null);

			if ( pb_deconnectRemote && debugProtocol.isActive() ) {
				new ExecutionTerminatedEventRequest( debugProtocol ).sendAsync();
			}
		}
	}
	
	public boolean isTerminated() {
		return debuggerState == EDebuggerState.TERMINATED;
	}

	public void connectToDebugClient(String clientAddress, int clientServerPort) throws InterruptedException
	{
		/*
		 * Open connection to client
		 */
		IConnector connector = TCPUtil.getConnector(this.debugProtocolContainer, clientAddress + ":" + clientServerPort);
		this.debugProtocol.open(connector);
	}

	public Stack<StackFrame<StoryDiagramElement>> getExecutionStack()
	{
		return this.executionStack;
	}

	public Stack<ChangeHistoryEntry> getChangeHistory()
	{
		return this.changeHistory;
	}

	protected void addChangeHistoryEntry(InterpreterNotification<EClassifier> interpreterNotification)
	{
		ChangeDescription changeDescription = this.changeRecorder.endRecording();

		this.getChangeHistory().push(new ChangeHistoryEntry(changeDescription, interpreterNotification));

		this.changeRecorder.beginRecording(Collections.singleton(this.debuggerResourceSet));
	}

	public SynchronizerAdapter getSynchronizerAdapter()
	{
		return this.synchronizerAdapter;
	}

	public int getResourceSetSynchronizerPort()
	{
		return this.resourceSetSynchronizerPort;
	}

	public ResourceSet getDebuggerResourceSet()
	{
		return this.debuggerResourceSet;
	}

	public Resource getDebuggerResource()
	{
		return this.debuggerResource;
	}

	public Map<String, Collection<Breakpoint<StoryDiagramElement>>> getBreakpoints()
	{
		return this.breakpoints;
	}

	public SDMInterpreter<Activity, ActivityNode, ?, StoryPattern, ?, ?, EClassifier, EStructuralFeature, Expression> getInterpreter()
	{
		return this.interpreter;
	}

	public void setProcessNotifications(boolean processNotifications)
	{
		this.processNotifications = processNotifications;
	}
	
	private boolean shouldProcessNoditication( final InterpreterNotification<EClassifier> p_notification  ) {
		if ( processNotifications ) {
			switch ( p_notification.getNotificationType() ) {
//				case STORY_PATTERN_INITIALIZATION_STARTED:
//					return false;
	
				default:
					return true;
			}
		}
		
		return false;
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public void notifyChanged( final InterpreterNotification<EClassifier> p_notification ) {
		if ( shouldProcessNoditication( p_notification ) ) {

			/*
			 * Initialize and suspend the interpreter if the execution of the
			 * root activity has just started (i.e. the execution stack contains
			 * only one element).
			 * DB: Issue 1825: avoid passing in the initialization more than once
			 */
			if ( getExecutionStack().size() == 1 && !isInitialized() /*&& p_notification.getNotificationType() == NotificationTypeEnum.ACTIVITY_EXECUTION_STARTED*/ ) {
				// DB: Issue 1825: The interpreter is already set in the constructor
				//assert p_notification.getNotifier() instanceof SDMInterpreter;
				
				final Object sdElement = p_notification.getMainStoryDiagramElement();
				
				assert sdElement instanceof EObject;
				
				final Resource resource = ( (EObject) sdElement ).eResource();
				
				assert resource != null;

				initialize( resource.getResourceSet()/*,
							(SDMInterpreter<Activity, ActivityNode, ?, StoryPattern, ?, ?, EClassifier, EStructuralFeature, Expression>) p_notification.getNotifier()*/ );
//				initialize( ((ActivityExecutionStartedNotification<Activity, ?>) p_notification ).getActivity(),
//						(SDMInterpreter<Activity, ActivityNode, ?, StoryPattern, ?, ?, EClassifier, EStructuralFeature, Expression>) p_notification.getNotifier() );

				synchronized (this)	{
					try	{
						System.out.println( "Interpreter waiting for being resumed by client..." );
						this.wait();
					}
					catch ( final Throwable p_th ) {
						p_th.printStackTrace();
						
						throw new RuntimeException( p_th );
					}
				}
			}

			/*
			 * add a new stack frame if the last notification of the current top
			 * stack frame is one of the following types
			 */
			final InterpreterNotification<EClassifier> lastNotification = getExecutionStack().peek().getLastNotification();

			if ( lastNotification != null ) {
				switch ( lastNotification.getNotificationType() ) {
					case ACTIVITY_EXECUTION_STARTED:
					case ACTIVITY_NODE_EXECUTION_STARTED:
					case EVALUATING_EXPRESSION:
					case STORY_PATTERN_INITIALIZATION_STARTED:
					case STORY_PATTERN_MATCHING_STARTED:
					case STORY_PATTERN_APPLICATION_STARTED:
						getExecutionStack().push( new StackFrame<StoryDiagramElement>() );
						
						break;

					default:
						break;
				}
			}

			/*
			 * add entry to change history
			 */
			this.addChangeHistoryEntry( p_notification );

			/*
			 * remove the top stack frame if the current notification is of the
			 * following types
			 */
			switch ( p_notification.getNotificationType() )	{
				case ACTIVITY_EXECUTION_FINISHED:
				case ACTIVITY_NODE_EXECUTION_FINISHED:
				case EVALUATED_EXPRESSION:
				case STORY_PATTERN_INITIALIZATION_FINISHED:
				case STORY_PATTERN_MATCHING_SUCCESSFUL:
				case STORY_PATTERN_MATCHING_FAILED:
				case STORY_PATTERN_APPLICATION_FINISHED:
					getExecutionStack().pop();
					
					break;

				default:
					break;
			}

			/*
			 * modify the top stack frame if the notification is connected to a
			 * story diagram element.
			 */
			if ( p_notification.getMainStoryDiagramElement() != null ) {
				final StackFrame<StoryDiagramElement> frame = this.getExecutionStack().peek();
				frame.setStoryDiagramElement((StoryDiagramElement) p_notification.getMainStoryDiagramElement());
				frame.setVariablesScope( p_notification.getVariablesScope( ));
				frame.setLastNotification( p_notification );
			}

			switch ( p_notification.getNotificationType() )	{

				/*
				 * Add created objects to the debugger resource
				 */
				case INSTANCE_OBJECT_CREATED:
					InstanceObjectCreatedNotification<?, ?> icn = (InstanceObjectCreatedNotification<?, ?>) p_notification;

					if (icn.getInstanceObject() instanceof EObject && ((EObject) icn.getInstanceObject()).eContainer() == null)	{
						this.getDebuggerResource().getContents().add((EObject) icn.getInstanceObject());
					}

					break;

				/*
				 * add objects to the debugger resource if instance links are
				 * deleted and the connected objects have no container.
				 */
				case INSTANCE_LINK_DESTROYED:
					InstanceLinkDestroyedNotification<?, ?, ?> ildn = (InstanceLinkDestroyedNotification<?, ?, ?>) p_notification;

					if (ildn.getSourceObject() instanceof EObject && ((EObject) ildn.getSourceObject()).eContainer() == null) {
						this.getDebuggerResource().getContents().add((EObject) ildn.getSourceObject());
					}

					if (ildn.getTargetObject() instanceof EObject && ((EObject) ildn.getTargetObject()).eContainer() == null) {
						this.getDebuggerResource().getContents().add((EObject) ildn.getTargetObject());
					}

					break;

				default:
					break;
			}

			/*
			 * If the stack size is 1 and the execution of an activity has
			 * finished, emit an appropriate debug event.
			 */
			if ( getExecutionStack().size() == 1 && p_notification.getNotificationType() == NotificationTypeEnum.ACTIVITY_EXECUTION_FINISHED ) {
				try	{
					if ( debugProtocol.isActive() )	{
						new ExecutionFinishedEventRequest( debugProtocol ).sendAsync();
					}
				}
				catch (final Throwable p_th )	{
					p_th.printStackTrace();
					
					throw new RuntimeException( p_th );
				}
			}
			/*
			 * Check if the conditions of break points are satisfied and suspend
			 * if necessary.
			 */
			else if ( p_notification.getMainStoryDiagramElement() != null ) {
				suspendOnBreakpoints( (StoryDiagramElement) p_notification.getMainStoryDiagramElement(), p_notification.getVariablesScope() );
			}
		}
	}

	public void abort(SDMInterpreter<?, ?, ?, ?, ?, ?, ?, ?, ?> targetInterpreter)
	throws Exception {
		if ( getDebuggerState() != EDebuggerState.TERMINATED ) {
			synchronized ( this )	{
				System.out.println("---TERMINATING");

				debuggerState = EDebuggerState.TERMINATED;

				abortException = new AbortInterpretationException(targetInterpreter);

				notify();
			}
		}
	}
}
